# SPDX-FileCopyrightText: 2018-2021 The glTF-Blender-IO authors
#
# SPDX-License-Identifier: Apache-2.0

import base64

from ...io.com import gltf2_io
from ...io.exp import binary_data as gltf2_io_binary_data


class Buffer:
    """Class representing binary data for use in a glTF file as 'buffer' property."""

    def __init__(self, buffer_index=0, initial_data=None):
        self.__data = bytearray(b"")
        if initial_data is not None:
            self.__data = bytearray(initial_data.tobytes())
        self.__buffer_index = buffer_index

    def add_and_get_view(self, binary_data: gltf2_io_binary_data.BinaryData) -> gltf2_io.BufferView:
        """Add binary data to the buffer. Return a glTF BufferView."""
        offset = len(self.__data)
        self.__data.extend(binary_data.data)

        length = binary_data.byte_length

        # offsets should be a multiple of 4 --> therefore add padding if necessary
        padding = (4 - (length % 4)) % 4
        self.__data.extend(b"\x00" * padding)

        buffer_view = gltf2_io.BufferView(
            buffer=self.__buffer_index,
            byte_length=length,
            byte_offset=offset,
            byte_stride=None,
            extensions=None,
            extras=None,
            name=None,
            target=binary_data.bufferViewTarget
        )
        return buffer_view

    @property
    def byte_length(self):
        return len(self.__data)

    def to_bytes(self):
        return self.__data

    def clear(self):
        self.__data = b""

    def to_embed_string(self):
        return 'data:application/octet-stream;base64,' + base64.b64encode(self.__data).decode('ascii')



# SPDX-FileCopyrightText: 2018-2023 The glTF-Blender-IO authors
#
# SPDX-License-Identifier: Apache-2.0

import bpy
import typing
from ......io.com import gltf2_io
from ......io.exp import binary_data as gltf2_io_binary_data
from ......io.com import constants as gltf2_io_constants
from ....cache import cached
from ....accessors import gather_accessor
from .keyframes import gather_data_sampled_keyframes


@cached
def gather_data_sampled_animation_sampler(
        blender_type_data: str,
        blender_id: str,
        channel: str,
        action_name: str,
        slot_identifier: str,
        node_channel_is_animated: bool,
        node_channel_interpolation: str,
        additional_key: str,  # Used to differentiate between material / material node_tree
        export_settings
):

    keyframes = __gather_keyframes(
        blender_type_data,
        blender_id,
        channel,
        action_name,
        slot_identifier,
        node_channel_is_animated,
        additional_key,
        export_settings)

    if keyframes is None:
        # After check, no need to animate this node for this channel
        return None

    # Now we are raw input/output, we need to convert to glTF data
    input, output = __convert_keyframes(blender_type_data, blender_id, channel, keyframes, action_name, export_settings)

    sampler = gltf2_io.AnimationSampler(extensions=None, extras=None, input=input, interpolation=__gather_interpolation(
        blender_type_data, node_channel_is_animated, node_channel_interpolation, keyframes, export_settings), output=output)

    return sampler


def __gather_keyframes(
        blender_type_data,
        blender_id,
        channel,
        action_name,
        slot_identifier,
        node_channel_is_animated,
        additional_key,  # Used to differentiate between material / material node_tree
        export_settings):

    keyframes = gather_data_sampled_keyframes(
        blender_type_data,
        blender_id,
        channel,
        action_name,
        slot_identifier,
        node_channel_is_animated,
        additional_key,
        export_settings
    )

    if keyframes is None:
        # After check, no need to animation this node
        return None

    return keyframes


def __convert_keyframes(blender_type_data, blender_id, channel, keyframes, action_name, export_settings):

    # Sliding can come from:
    # - option SLIDE for negative frames
    # - option to start animation at frame 0 for looping
    if blender_id in export_settings['slide'].keys() and action_name in export_settings['slide'][blender_id].keys():
        for k in keyframes:
            k.frame += -export_settings['slide'][blender_id][action_name]
            k.seconds = k.frame / bpy.context.scene.render.fps

    times = [k.seconds for k in keyframes]
    input = gather_accessor(
        gltf2_io_binary_data.BinaryData.from_list(times, gltf2_io_constants.ComponentType.Float),
        gltf2_io_constants.ComponentType.Float,
        len(times),
        tuple([max(times)]),
        tuple([min(times)]),
        gltf2_io_constants.DataType.Scalar,
        export_settings)

    values = []
    for keyframe in keyframes:
        keyframe_value = __convert_to_gltf(keyframe.value)
        values += keyframe_value

     # store the keyframe data in a binary buffer
    component_type = gltf2_io_constants.ComponentType.Float
    if type(keyframes[0].value).__name__ != "float":
        data_type = gltf2_io_constants.DataType.vec_type_from_num(len(keyframes[0].value))
    else:
        data_type = gltf2_io_constants.DataType.vec_type_from_num(1)

    output = gather_accessor(
        gltf2_io_binary_data.BinaryData.from_list(values, component_type),
        component_type,
        len(values) // gltf2_io_constants.DataType.num_elements(data_type),
        None,
        None,
        data_type,
        export_settings
    )

    return input, output


def __gather_interpolation(
        blender_type_data,
        node_channel_is_animated,
        node_channel_interpolation,
        keyframes,
        export_settings):
    # TODOPointer
    return export_settings['gltf_sampling_interpolation_fallback']


def __convert_to_gltf(value):
    return value if type(value).__name__ != "float" else [value]



# SPDX-FileCopyrightText: 2018-2021 The glTF-Blender-IO authors
#
# SPDX-License-Identifier: Apache-2.0

import mathutils
from ...io.com import gltf2_io, constants as gltf2_io_constants
from ...io.exp import binary_data as gltf2_io_binary_data
from ...io.exp.user_extensions import export_user_extensions
from . import accessors as gltf2_blender_gather_accessors
from . import joints as gltf2_blender_gather_joints
from .tree import VExportNode
from .cache import cached


@cached
def gather_skin(armature_uuid, export_settings):
    """
    Gather armatures, bones etc into a glTF2 skin object.

    :param blender_object: the object which may contain a skin
    :param export_settings:
    :return: a glTF2 skin object
    """

    if armature_uuid not in export_settings['vtree'].nodes:
        # User filtered objects to export, and keep the skined mesh, without keeping the armature
        return None

    blender_armature_object = export_settings['vtree'].nodes[armature_uuid].blender_object

    if not __filter_skin(blender_armature_object, export_settings):
        return None

    skin = gltf2_io.Skin(
        extensions=__gather_extensions(blender_armature_object, export_settings),
        extras=__gather_extras(blender_armature_object, export_settings),
        inverse_bind_matrices=__gather_inverse_bind_matrices(armature_uuid, export_settings),
        joints=__gather_joints(armature_uuid, export_settings),
        name=__gather_name(blender_armature_object, export_settings),
        skeleton=__gather_skeleton(blender_armature_object, export_settings)
    )

    # If armature is not exported, joints will be empty.
    # Do not construct skin in that case
    if len(skin.joints) == 0:
        return None

    export_user_extensions('gather_skin_hook', export_settings, skin, blender_armature_object)

    return skin


def __filter_skin(blender_armature_object, export_settings):
    if not export_settings['gltf_skins']:
        return False
    if blender_armature_object.type != 'ARMATURE' or len(blender_armature_object.pose.bones) == 0:
        return False

    return True


def __gather_extensions(blender_armature_object, export_settings):
    return None


def __gather_extras(blender_armature_object, export_settings):
    return None


def __gather_inverse_bind_matrices(armature_uuid, export_settings):

    blender_armature_object = export_settings['vtree'].nodes[armature_uuid].blender_object

    axis_basis_change = mathutils.Matrix.Identity(4)
    if export_settings['gltf_yup']:
        axis_basis_change = mathutils.Matrix(
            ((1.0, 0.0, 0.0, 0.0), (0.0, 0.0, 1.0, 0.0), (0.0, -1.0, 0.0, 0.0), (0.0, 0.0, 0.0, 1.0)))

    # store matrix_world of armature in case we need to add a neutral bone
    export_settings['vtree'].nodes[armature_uuid].matrix_world_armature = blender_armature_object.matrix_world.copy()

    bones_uuid = export_settings['vtree'].get_all_bones(armature_uuid)

    def __collect_matrices(bone):
        inverse_bind_matrix = (
            axis_basis_change @
            (
                blender_armature_object.matrix_world @
                bone.bone.matrix_local
            )
        ).inverted_safe()
        matrices.append(inverse_bind_matrix)

    matrices = []
    for b in bones_uuid:
        if export_settings['vtree'].nodes[b].leaf_reference is None:
            __collect_matrices(blender_armature_object.pose.bones[export_settings['vtree'].nodes[b].blender_bone.name])
        else:
            inverse_bind_matrix = (
                axis_basis_change @
                (
                    blender_armature_object.matrix_world @
                    export_settings['vtree'].nodes[export_settings['vtree'].nodes[b].leaf_reference].matrix_world_tail
                )
            ).inverted_safe()
            matrices.append(inverse_bind_matrix)  # Leaf bone

    # flatten the matrices
    inverse_matrices = []
    for matrix in matrices:
        for column in range(0, 4):
            for row in range(0, 4):
                inverse_matrices.append(matrix[row][column])

    binary_data = gltf2_io_binary_data.BinaryData.from_list(inverse_matrices, gltf2_io_constants.ComponentType.Float)
    return gltf2_blender_gather_accessors.gather_accessor(
        binary_data,
        gltf2_io_constants.ComponentType.Float,
        len(inverse_matrices) // gltf2_io_constants.DataType.num_elements(gltf2_io_constants.DataType.Mat4),
        None,
        None,
        gltf2_io_constants.DataType.Mat4,
        export_settings
    )


def __gather_joints(armature_uuid, export_settings):

    all_armature_children = export_settings['vtree'].nodes[armature_uuid].children
    root_bones_uuid = [
        c for c in all_armature_children if export_settings['vtree'].nodes[c].blender_type == VExportNode.BONE]

    # Create bone nodes
    for root_bone_uuid in root_bones_uuid:
        gltf2_blender_gather_joints.gather_joint_vnode(root_bone_uuid, export_settings)

    bones_uuid = export_settings['vtree'].get_all_bones(armature_uuid)
    joints = [export_settings['vtree'].nodes[b].node for b in bones_uuid]
    return joints


def __gather_name(blender_armature_object, export_settings):
    return blender_armature_object.name


def __gather_skeleton(blender_armature_object, export_settings):
    # In the future support the result of https://github.com/KhronosGroup/glTF/pull/1195
    return None



# SPDX-FileCopyrightText: 2018-2023 The glTF-Blender-IO authors
#
# SPDX-License-Identifier: Apache-2.0

from ...io.com.gltf2_io import TextureInfo
from .texture import texture
from ..com.conversion import get_anisotropy_rotation_gltf_to_blender
from math import pi
from mathutils import Vector


def anisotropy(
        mh,
        location,
        anisotropy_socket,
        anisotropy_rotation_socket,
        anisotropy_tangent_socket
):

    if anisotropy_socket is None or anisotropy_rotation_socket is None or anisotropy_tangent_socket is None:
        return

    x, y = location
    try:
        ext = mh.pymat.extensions['KHR_materials_anisotropy']
        # Needed for KHR_animation_pointer
        mh.pymat.extensions['KHR_materials_anisotropy']['blender_nodetree'] = mh.node_tree
        mh.pymat.extensions['KHR_materials_anisotropy']['blender_mat'] = mh.mat  # Needed for KHR_animation_pointer
    except Exception:
        return

    anisotropy_strength = ext.get('anisotropyStrength', 0)
    anisotropy_rotation = ext.get('anisotropyRotation', 0)
    tex_info = ext.get('anisotropyTexture')
    if tex_info is not None:
        tex_info = TextureInfo.from_dict(tex_info)

    # We are going to use UVMap of Normal map if it exists, as input for the anisotropy tangent

    if tex_info is None:
        anisotropy_socket.default_value = anisotropy_strength
        anisotropy_rotation_socket.default_value = get_anisotropy_rotation_gltf_to_blender(anisotropy_rotation)
        return

    # Tangent node
    node = mh.node_tree.nodes.new('ShaderNodeTangent')
    node.direction_type = "UV_MAP"
    node.location = x - 180, y - 200
    uv_idx = tex_info.tex_coord or 0

    # Get the UVMap of the normal map if available (if not, keeping the first UVMap available, uv_idx = 0)
    tex_info_normal = mh.pymat.normal_texture
    if tex_info_normal is not None:
        try:
            uv_idx = tex_info.extensions['KHR_texture_transform']['texCoord']
        except Exception:
            pass

    node.uv_map = 'UVMap' if uv_idx == 0 else 'UVMap.%03d' % uv_idx
    mh.node_tree.links.new(anisotropy_tangent_socket, node.outputs['Tangent'])

    # Multiply node
    multiply_node = mh.node_tree.nodes.new('ShaderNodeMath')
    multiply_node.label = 'Anisotropy strength'
    multiply_node.operation = 'MULTIPLY'
    multiply_node.location = x - 180, y + 200
    mh.node_tree.links.new(anisotropy_socket, multiply_node.outputs[0])
    multiply_node.inputs[1].default_value = anisotropy_strength

    # Divide node
    divide_node = mh.node_tree.nodes.new('ShaderNodeMath')
    divide_node.label = 'Rotation conversion'
    divide_node.operation = 'DIVIDE'
    divide_node.location = x - 180, y
    mh.node_tree.links.new(anisotropy_rotation_socket, divide_node.outputs[0])
    divide_node.inputs[1].default_value = 2 * pi

    # Rotation node
    rotation_node = mh.node_tree.nodes.new('ShaderNodeMath')
    rotation_node.label = 'Anisotropy rotation'
    rotation_node.operation = 'ADD'
    rotation_node.location = x - 180 * 2, y
    mh.node_tree.links.new(divide_node.inputs[0], rotation_node.outputs[0])
    rotation_node.inputs[1].default_value = anisotropy_rotation

    # ArcTan node
    arctan_node = mh.node_tree.nodes.new('ShaderNodeMath')
    arctan_node.label = 'ArcTan2'
    arctan_node.operation = 'ARCTAN2'
    arctan_node.location = x - 180 * 3, y
    mh.node_tree.links.new(rotation_node.inputs[0], arctan_node.outputs[0])

    # Separate XYZ
    sep_node = mh.node_tree.nodes.new('ShaderNodeSeparateXYZ')
    sep_node.location = x - 180 * 4, y
    mh.node_tree.links.new(arctan_node.inputs[0], sep_node.outputs[1])
    mh.node_tree.links.new(arctan_node.inputs[1], sep_node.outputs[0])
    mh.node_tree.links.new(multiply_node.inputs[0], sep_node.outputs[2])

    # Multiply add node
    multiply_add_node = mh.node_tree.nodes.new('ShaderNodeVectorMath')
    multiply_add_node.location = x - 180 * 5, y
    multiply_add_node.operation = 'MULTIPLY_ADD'
    multiply_add_node.inputs[1].default_value = Vector((2, 2, 1))
    multiply_add_node.inputs[2].default_value = Vector((-1, -1, 0))
    mh.node_tree.links.new(sep_node.inputs[0], multiply_add_node.outputs[0])

    # Texture
    texture(
        mh,
        tex_info=tex_info,
        label='ANISOTROPY',
        location=(x - 180 * 6, y),
        is_data=True,
        color_socket=multiply_add_node.inputs[0]
    )



# SPDX-FileCopyrightText: 2018-2022 The glTF-Blender-IO authors
#
# SPDX-License-Identifier: Apache-2.0

import bpy
import typing
from ......io.com import gltf2_io
from ......io.exp.user_extensions import export_user_extensions
from .....com.conversion import get_gltf_interpolation
from .....com.conversion import get_target, get_channel_from_target
from ...fcurves.channels import get_channel_groups
from ...fcurves.channels import needs_baking
from ...drivers import get_sk_drivers
from ..object.channels import gather_sampled_object_channel
from ..shapekeys.channels import gather_sampled_sk_channel
from .channel_target import gather_armature_sampled_channel_target
from .sampler import gather_bone_sampled_animation_sampler


def gather_armature_sampled_channels(armature_uuid, blender_action_name, slot_identifier,
                                     export_settings) -> typing.List[gltf2_io.AnimationChannel]:
    channels = []
    extra_channels = {}

    # Then bake all bones
    bones_to_be_animated = []
    bones_uuid = export_settings["vtree"].get_all_bones(armature_uuid)
    bones_to_be_animated = [
        export_settings["vtree"].nodes[b].blender_bone.name for b in bones_uuid if export_settings["vtree"].nodes[b].leaf_reference is None]

    # List of really animated bones is needed for optimization decision
    list_of_animated_bone_channels = {}
    if slot_identifier is not None:
        if armature_uuid != blender_action_name and blender_action_name in bpy.data.actions:
            # Not bake situation
            channels_animated, to_be_sampled, extra_channels = get_channel_groups(
                armature_uuid, bpy.data.actions[blender_action_name],
                    bpy.data.actions[blender_action_name].slots[slot_identifier], export_settings)
            for chan in [chan for chan in channels_animated.values() if chan['bone'] is not None]:
                for prop in chan['properties'].keys():
                    list_of_animated_bone_channels[(chan['bone'], get_channel_from_target(get_target(prop)))] = get_gltf_interpolation(
                        chan['properties'][prop][0].keyframe_points[0].interpolation, export_settings)  # Could be exported without sampling : keep interpolation

            for _, _, chan_prop, chan_bone in [chan for chan in to_be_sampled if chan[1] == "BONE"]:
                list_of_animated_bone_channels[
                    (
                        chan_bone,
                        chan_prop,
                    )
                ] = get_gltf_interpolation(export_settings['gltf_sampling_interpolation_fallback'], export_settings)  # if forced to be sampled, keep the interpolation chosen by the user
    else:
        pass
        # There is no animated channels (because if it was, we would have a slot_identifier)
        # We are in a bake situation

    for bone in bones_to_be_animated:
        for p in ["location", "rotation_quaternion", "scale"]:
            channel = gather_sampled_bone_channel(
                armature_uuid,
                bone,
                p,
                blender_action_name,
                slot_identifier,
                (bone, p) in list_of_animated_bone_channels.keys(),
                list_of_animated_bone_channels[(bone, p)] if (bone, p) in list_of_animated_bone_channels.keys() else get_gltf_interpolation(export_settings['gltf_sampling_interpolation_fallback'], export_settings),
                export_settings)
            if channel is not None:
                channels.append(channel)

    bake_interpolation = get_gltf_interpolation(export_settings['gltf_sampling_interpolation_fallback'], export_settings)
    # Retrieve animation on armature object itself, if any
    if blender_action_name == armature_uuid or export_settings['gltf_animation_mode'] in ["SCENE", "NLA_TRACKS"]:
        # If armature is baked (no animation of armature), need to use all channels
        armature_channels = [
            ["location", bake_interpolation],
            ["rotation_quaternion", bake_interpolation],
            ["scale", bake_interpolation]
        ]
        animated_channels = []
    else:
        # The armature has some channel(s) animated, checking which one(s)
        armature_channels = __gather_armature_object_channel(
            armature_uuid, bpy.data.actions[blender_action_name], slot_identifier, export_settings)
        animated_channels = armature_channels

    for (p, i) in armature_channels:
        armature_channel = gather_sampled_object_channel(
            armature_uuid,
            p,
            blender_action_name,
            slot_identifier,
            p in [a[0] for a in animated_channels],
            [c[1] for c in animated_channels if c[0] == p][0] if p in [a[0] for a in animated_channels] else bake_interpolation,
            export_settings
        )

        if armature_channel is not None:
            channels.append(armature_channel)

    # Retrieve channels for drivers, if needed
    drivers_to_manage = get_sk_drivers(armature_uuid, export_settings)
    for obj_driver_uuid in drivers_to_manage:
        channel = gather_sampled_sk_channel(obj_driver_uuid, armature_uuid + "_" + blender_action_name, slot_identifier, export_settings)
        if channel is not None:
            channels.append(channel)

    return channels, extra_channels


def gather_sampled_bone_channel(
        armature_uuid: str,
        bone: str,
        channel: str,
        action_name: str,
        slot_identifier: str,
        node_channel_is_animated: bool,
        node_channel_interpolation: str,
        export_settings
):

    __target = __gather_target(armature_uuid, bone, channel, export_settings)
    if __target.path is not None:
        sampler = __gather_sampler(
            armature_uuid,
            bone,
            channel,
            action_name,
            slot_identifier,
            node_channel_is_animated,
            node_channel_interpolation,
            export_settings)

        if sampler is None:
            # After check, no need to animate this node for this channel
            return None

        animation_channel = gltf2_io.AnimationChannel(
            extensions=None,
            extras=None,
            sampler=sampler,
            target=__target
        )

        export_user_extensions('gather_animation_channel_hook',
                               export_settings,
                               animation_channel,
                               channel,
                               export_settings['vtree'].nodes[armature_uuid].blender_object,
                               bone,
                               action_name,
                               node_channel_is_animated
                               )

        return animation_channel
    return None


def __gather_target(armature_uuid: str,
                    bone: str,
                    channel: str,
                    export_settings
                    ) -> gltf2_io.AnimationChannelTarget:

    return gather_armature_sampled_channel_target(
        armature_uuid, bone, channel, export_settings)


def __gather_sampler(
        armature_uuid,
        bone,
        channel,
        action_name,
        slot_identifier, #TODOSLOT
        node_channel_is_animated,
        node_channel_interpolation,
        export_settings):
    return gather_bone_sampled_animation_sampler(
        armature_uuid,
        bone,
        channel,
        action_name,
        slot_identifier, #TODOSLOT
        node_channel_is_animated,
        node_channel_interpolation,
        export_settings
    )


def __gather_armature_object_channel(obj_uuid: str, blender_action, slot_identifier, export_settings):
    channels = []

    channels_animated, to_be_sampled, extra_channels = get_channel_groups(obj_uuid, blender_action, blender_action.slots[slot_identifier], export_settings)
    # Remove all channel linked to bones, keep only directly object channels
    channels_animated = [c for c in channels_animated.values() if c['type'] == "OBJECT"]
    to_be_sampled = [c for c in to_be_sampled if c[1] == "OBJECT"]

    original_channels = []
    for c in channels_animated:
        original_channels.extend([(prop, c['properties'][prop][0].keyframe_points[0].interpolation)
                                 for prop in c['properties'].keys()])

    for c, inter in original_channels:
        channels.append(
            (
                {
                    "location": "location",
                    "rotation_quaternion": "rotation_quaternion",
                    "rotation_euler": "rotation_quaternion",
                    "scale": "scale",
                    "delta_location": "location",
                    "delta_scale": "scale",
                    "delta_rotation_euler": "rotation_quaternion",
                    "delta_rotation_quaternion": "rotation_quaternion"
                }.get(c),
                get_gltf_interpolation(inter, export_settings)
            )
        )

    for c in to_be_sampled:
        channels.append(
            (
                {
                    "location": "location",
                    "rotation_quaternion": "rotation_quaternion",
                    "rotation_euler": "rotation_quaternion",
                    "scale": "scale",
                    "delta_location": "location",
                    "delta_scale": "scale",
                    "delta_rotation_euler": "rotation_quaternion",
                    "delta_rotation_quaternion": "rotation_quaternion"
                }.get(c[2]),
                get_gltf_interpolation(export_settings['gltf_sampling_interpolation_fallback'], export_settings)  # Forced to be sampled, so use the interpolation chosen by the user
            )
        )

    return channels



# SPDX-FileCopyrightText: 2018-2021 The glTF-Blender-IO authors
#
# SPDX-License-Identifier: Apache-2.0

from .pbrMetallicRoughness import base_color


def unlit(mh):
    """Creates node tree for unlit materials."""
    # Emission node for the base color
    emission_node = mh.nodes.new('ShaderNodeEmission')
    emission_node.location = 10, 126

    # Create a "Lightpath trick": makes Emission visible only to
    # camera rays, so it won't "glow" in Cycles.
    #
    # [Is Camera Ray] => [Mix] =>
    #   [Transparent] => [   ]
    #      [Emission] => [   ]
    lightpath_node = mh.nodes.new('ShaderNodeLightPath')
    transparent_node = mh.nodes.new('ShaderNodeBsdfTransparent')
    mix_node = mh.nodes.new('ShaderNodeMixShader')
    lightpath_node.location = 10, 600
    transparent_node.location = 10, 240
    mix_node.location = 260, 320
    mh.links.new(mix_node.inputs['Fac'], lightpath_node.outputs['Is Camera Ray'])
    mh.links.new(mix_node.inputs[1], transparent_node.outputs[0])
    mh.links.new(mix_node.inputs[2], emission_node.outputs[0])

    # Material output
    alpha_socket = None
    out_node = mh.nodes.new('ShaderNodeOutputMaterial')
    if mh.is_opaque():
        out_node.location = 490, 290
        mh.links.new(out_node.inputs[0], mix_node.outputs[0])
    else:
        # Create a "Mix with Transparent" setup so there's a
        # place to put Alpha.
        #
        #         Alpha => [Mix] => [Output]
        # [Transparent] => [   ]
        #         Color => [   ]
        mix2_node = mh.nodes.new('ShaderNodeMixShader')
        alpha_socket = mix2_node.inputs['Fac']
        mix2_node.location = 490, -50
        out_node.location = 700, -70
        mh.links.new(mix2_node.inputs[1], transparent_node.outputs[0])
        mh.links.new(mix2_node.inputs[2], mix_node.outputs[0])
        mh.links.new(out_node.inputs[0], mix2_node.outputs[0])

    base_color(
        mh,
        location=(-200, 380),
        color_socket=emission_node.inputs['Color'],
        alpha_socket=alpha_socket,
    )



import bpy
cycles = bpy.context.scene.cycles

cycles.use_preview_adaptive_sampling = True
cycles.preview_adaptive_threshold = 0.01
cycles.preview_samples = 4096
cycles.preview_adaptive_min_samples = 0
cycles.use_preview_denoising = True
cycles.preview_denoiser = 'OPENIMAGEDENOISE'
cycles.preview_denoising_input_passes = 'RGB_ALBEDO_NORMAL'
cycles.preview_denoising_prefilter = 'ACCURATE'



# SPDX-FileCopyrightText: 2008-2023 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

#  Author   : Stephane Grabli
#  Date     : 04/08/2005
#  Purpose  : Builds sketchy strokes whose topology relies on a
#             parameterization that covers the complete lines (visible+invisible)
#             whereas only the visible portions are actually drawn

from freestyle.chainingiterators import pySketchyChainSilhouetteIterator
from freestyle.predicates import (
    QuantitativeInvisibilityUP1D,
    TrueUP1D,
)
from freestyle.shaders import (
    IncreasingColorShader,
    IncreasingThicknessShader,
    SamplingShader,
    SmoothingShader,
    SpatialNoiseShader,
    pyHLRShader,
)
from freestyle.types import Operators


Operators.select(QuantitativeInvisibilityUP1D(0))
Operators.bidirectional_chain(pySketchyChainSilhouetteIterator(3, False))
shaders_list = [
    SamplingShader(2),
    SpatialNoiseShader(15, 120, 2, True, True),
    IncreasingThicknessShader(5, 30),
    SmoothingShader(100, 0.05, 0, 0.2, 0, 0, 0, 1),
    IncreasingColorShader(0, 0.2, 0, 1, 0.2, 0.7, 0.2, 1),
    pyHLRShader(),
]
Operators.create(TrueUP1D(), shaders_list)



# SPDX-FileCopyrightText: 2018-2024 The glTF-Blender-IO authors
#
# SPDX-License-Identifier: Apache-2.0

from ...com.extras import generate_extras

def gather_extras(blender_material, export_settings):
    if export_settings['gltf_extras']:
        return generate_extras(blender_material)
    return None

def gather_name(blender_material, export_settings):
    return blender_material.name



# SPDX-FileCopyrightText: 2018-2022 The glTF-Blender-IO authors
#
# SPDX-License-Identifier: Apache-2.0

import bpy
from ......io.com import gltf2_io, constants as gltf2_io_constants
from ......io.exp import binary_data as gltf2_io_binary_data
from ......io.exp.user_extensions import export_user_extensions
from .....com.gltf2_blender_math import mathutils_to_gltf
from ....accessors import gather_accessor
from .keyframes import gather_sk_sampled_keyframes


def gather_sk_sampled_animation_sampler(
        obj_uuid,
        action_name,
        slot_identifier,
        export_settings
):

    keyframes = __gather_keyframes(
        obj_uuid,
        action_name,
        slot_identifier,
        export_settings)

    if keyframes is None:
        # After check, no need to animate this node for this channel
        return None

    # Now we are raw input/output, we need to convert to glTF data
    input, output = __convert_keyframes(obj_uuid, keyframes, action_name, export_settings)

    sampler = gltf2_io.AnimationSampler(
        extensions=None,
        extras=None,
        input=input,
        interpolation=__gather_interpolation(export_settings),
        output=output
    )

    blender_object = export_settings['vtree'].nodes[obj_uuid].blender_object
    export_user_extensions('animation_gather_sk_channels', export_settings, blender_object, action_name)

    return sampler


def __gather_keyframes(
        obj_uuid,
        action_name,
        slot_identifier,
        export_settings):

    keyframes = gather_sk_sampled_keyframes(
        obj_uuid,
        action_name,
        slot_identifier,
        export_settings
    )

    if keyframes is None:
        # After check, no need to animation this node
        return None

    return keyframes


def __convert_keyframes(obj_uuid, keyframes, action_name: str, export_settings):

    # Sliding can come from:
    # - option SLIDE for negative frames
    # - option to start animation at frame 0 for looping
    if obj_uuid in export_settings['slide'].keys() and action_name in export_settings['slide'][obj_uuid].keys():
        for k in keyframes:
            k.frame += -export_settings['slide'][obj_uuid][action_name]
            k.seconds = k.frame / (bpy.context.scene.render.fps * bpy.context.scene.render.fps_base)

    times = [k.seconds for k in keyframes]
    input = gather_accessor(
        gltf2_io_binary_data.BinaryData.from_list(times, gltf2_io_constants.ComponentType.Float),
        gltf2_io_constants.ComponentType.Float,
        len(times),
        tuple([max(times)]),
        tuple([min(times)]),
        gltf2_io_constants.DataType.Scalar,
        export_settings)

    values = []
    for keyframe in keyframes:
        keyframe_value = mathutils_to_gltf(keyframe.value)
        values += keyframe_value

    component_type = gltf2_io_constants.ComponentType.Float
    data_type = gltf2_io_constants.DataType.Scalar

    output = gather_accessor(
        gltf2_io_binary_data.BinaryData.from_list(values, component_type),
        component_type,
        len(values) // gltf2_io_constants.DataType.num_elements(data_type),
        None,
        None,
        data_type,
        export_settings
    )

    return input, output


def __gather_interpolation(export_settings):
    # TODO: check if the SK was animated with CONSTANT
    return export_settings['gltf_sampling_interpolation_fallback']
